#include "xhex.h"
#include <iostream>
#include <fstream>
#include <algorithm>
#include <boost/lexical_cast.hpp>
#include <boost/algorithm/string/case_conv.hpp>
#include "stdplatform.h"

static unsigned int charhex(char c) {
  unsigned char c2 = c;
  if (c2<='9') return (c2-'0');
  if (c2<='F') return (c2-'A'+10);
  if (c2<='f') return (c2-'a'+10);
  return 0;
}

int checkIntelHex(const std::string s) {
  int byteSize=0;
  int state = 0;
  int line = 1;
  int pos_bs = 0;
  int act_bs = 0;
  std::string::const_iterator it(s.begin()), end(s.end());
  while (it!=end) {
    char c = *it;
    if (state==0) {
      if (c==':') {
        state=1;
        pos_bs = 0;
      } else if (c=='\n') {
        state=0;
        line++;
      } else if (c=='\r') {
        state=0;
      } else {
        throw std::runtime_error("Syntax error in Intel hex data. Colon expected at begin of line!");
      }
    } else if (state==1) {
      pos_bs++;
      if (((c>='0')&&(c<='9')) || ((c>='a')&&(c<='f')) || ((c>='A')&&(c<='F'))) {
        state=1;
        if (pos_bs==1) {
          act_bs = charhex(c)<<4;
        } else if (pos_bs==2) {
          act_bs += charhex(c);
        } else if (pos_bs==7) {
          if (c!='0') pos_bs=9;
        } else if (pos_bs==8) {
          if (c=='0') byteSize+=act_bs;
        }
      } else if (c=='\n') {
        state=0;
        line++;
      } else if (c=='\r') {
        state=0;
      } else {
        throw std::runtime_error("Syntax error in Intel hex data. Hexadecimal digit expected in line!");
      }
    }
    it++;
  }
  return byteSize;
}

int checkSRec(const std::string s) {
  int byteSize=0;
  int state = 0;
  int line = 1;
  std::string::const_iterator it(s.begin()), end(s.end());
  while (it!=end) {
    char c = *it;
    if (state==0) {
      if (c=='S') {
        state=1;
      } else if (c=='\n') {
        state=0;
        line++;
      } else if (c=='\r') {
        state=0;
      } else {
        throw std::runtime_error("Syntax error in Motorola S-record data. 'S' expected at begin of line!");
      }
    } else if (state==1) {
      if (((c>='0')&&(c<='9')) || ((c>='a')&&(c<='f')) || ((c>='A')&&(c<='F'))) {
        state=1;
      } else if (c=='\n') {
        state=0;
        line++;
      } else if (c=='\r') {
        state=0;
      } else {
        throw std::runtime_error("Syntax error in Motorola S-record data. Hexadecimal digit expected in line!");
      }
    }
    it++;
  }
  return byteSize;
}

using namespace nicai;

xhex * xhex::create(const std::string & filename, const std::string & defPlatform) {
  if (filename.substr(filename.size()-5, 5) == ".srec") {
    return createFromSRec(filename, defPlatform);
  }
  if (filename.substr(filename.size()-4, 4) == ".hex") {
    return createFromHex(filename, defPlatform);
  }
  if (filename.substr(filename.size()-5, 5) == ".xhex") {
    return createFromXHex(filename);
  }
  if (filename.substr(filename.size()-5, 5) == ".bob3") {
    return createFromXHex(filename);
  }
  return 0;
}

xhex * xhex::createFromXHex(const std::string & filename) {
  std::ifstream in(filename.c_str());
  if (!in) {
    throw std::runtime_error("unable to open file: "+filename);
  }
  return createFromStream(in);
}

xhex * xhex::createFromStream(std::istream & in) {
  nicai::xml::ixmlstream xin(in);
  xhex * input_xhex = 0;
  xin.ignoreComments(true);
  xin.collapseWhitespace(true);

  if (xin.expectNode(nicai::xml::document())) {
    if (xin.expectNode(nicai::xml::tag("xhex"))) {
      input_xhex = new xhex(xin);
      xin.expectNode(nicai::xml::endtag());
    }
    xin.expectNode(nicai::xml::enddocument());
  } else {
    throw std::runtime_error("no xml document detected");
  }
  return input_xhex;
}

xhex * xhex::createFromHex(const std::string & filename, const std::string & defPlatform) {
  std::ifstream in(filename.c_str());
  if (!in) {
    throw std::runtime_error("unable to open file: "+filename);
  }


  std::string dummy ("<?xml version=\"1.0\"?>"
                     "<xhex version=\"1.0\">"
                     "<platform>Custom Intel-HEX</platform>"
                     //"<programmer type=\"usbasp\"/>"
                     "<device part=\"_\" erase=\"yes\">"
                     "<segment id=\"flash\" format=\"ihex\">"
                     "</segment>"
                     "</device>"
                     "</xhex>"
                     );
  std::istringstream sin(getStdPlatformXML(defPlatform));

  xhex * result = createFromStream(sin);

  // insert hex file into dummy xhex
  segment & seg = result->devicePtr->segmentVec[0];
  std::ostringstream s;
  s << in.rdbuf();
  seg.content = s.str();
  seg.byteSize = checkIntelHex(seg.content);
  return result;
}

xhex * xhex::createFromSRec(const std::string & filename, const std::string & defPlatform) {
  std::ifstream in(filename.c_str());
  if (!in) {
    throw std::runtime_error("unable to open file: "+filename);
  }

  std::string dummy ("<?xml version=\"1.0\"?>"
                     "<xhex version=\"1.0\">"
                     "<platform>Custom S-record</platform>"
                     //"<programmer type=\"usbasp\"/>"
                     "<device part=\"_\" erase=\"yes\">"
                     "<segment id=\"flash\" format=\"srec\">"
                     "</segment>"
                     "</device>"
                     "</xhex>"
                     );
  std::istringstream sin(getStdPlatformXML(defPlatform));

  xhex * result = createFromStream(sin);

  // insert hex file into dummy xhex
  segment & seg = result->devicePtr->segmentVec[0];
  std::ostringstream s;
  s << in.rdbuf();
  seg.content = s.str();
  checkSRec(seg.content);
  return result;
}


xhex::xhex(xml::ixmlstream & xin) {
  xml::attributemap attributes;
  xin.readAttributes(attributes);
  if (attributes["version"]!="1.0") {
    throw std::runtime_error("wrong file version");
  }

  while(!xin.expectNode(xml::endtag())) {
    if (xin.expectNode(xml::tag("platform"))) {
      platformPtr.reset(new platform(xin));
    } else if (xin.expectNode(xml::tag("programmer"))) {
      programmerPtr.reset(new programmer(xin));
    } else if (xin.expectNode(xml::tag("info"))) {
      infoVec.push_back(new info(xin));
    } else if (xin.expectNode(xml::tag("device"))) {
      devicePtr.reset(new device(xin));
    } else {
      xin.ignoreContent();
    }
  }
}


info * xhex::getInfo(const std::string & id) {
  for (unsigned int i=0; i<infoVec.size(); i++) {
    if (infoVec[i].id==id) {
      return &(infoVec[i]);
    }
  }
  return 0;
}

info::info(xml::ixmlstream & xin) {
  xml::attributemap attributes;
  xin.readAttributes(attributes);
  id = attributes["id"];
  xin.readText(content);
  xin.expectNode(xml::endtag());
}


platform::platform(xml::ixmlstream & xin) {
  xml::attributemap attributes;
  xin.readAttributes(attributes);
  xin.readText(content);
  xin.expectNode(xml::endtag());
}


programmer::programmer(xml::ixmlstream & xin) {
  xml::attributemap attributes;
  xin.readAttributes(attributes);
  type = attributes["type"];
  baudrate = attributes["baudrate"];
  bitclock = attributes["bitclock"];
  xin.expectNode(xml::endtag());
}


device::device(xml::ixmlstream & xin) {
  xml::attributemap attributes;
  xin.readAttributes(attributes);
  part = attributes["part"];
  erase = attributes["erase"]=="yes";

  while(!xin.expectNode(xml::endtag())) {
    if (xin.expectNode(xml::tag("segment"))) {
      segmentVec.push_back(new segment(xin));
    } else if (xin.expectNode(xml::tag("checksum"))) {
      //checksumVec.push_back(new checksum(xin));
    } else {
      xin.ignoreContent();
    }
  }
}

segment * device::getSegment(const std::string & id) {
  for (unsigned int i=0; i<segmentVec.size(); i++) {
    if (segmentVec[i].id==id) {
      return &(segmentVec[i]);
    }
  }
  return 0;
}

segment::segment(xml::ixmlstream & xin) {
  byteSize = 0;
  xml::attributemap attributes;
  xin.readAttributes(attributes);
  id = attributes["id"];
  format = attributes["format"];

  //std::cout << "Segment detected: " << id << std::endl;

  if (format=="") {
    while(!xin.expectNode(xml::endtag())) {
      if (xin.expectNode(xml::tag("data"))) {
        dataVec.push_back(new data(xin));
      } else if (xin.expectNode(xml::tag("checksum"))) {
        //checksumVec.push_back(new checksum(xin));
      } else {
        xin.ignoreContent();
      }
    }
  } else {
    xin.readText(content);
    std::replace(content.begin(), content.end(), ' ', '\n');
    content+='\n';
    if (format=="ihex") {
      byteSize=checkIntelHex(content);
    } else if (format=="srec") {
      byteSize=checkSRec(content);
    }
    xin.expectNode(xml::endtag());
  }
}


data::data(xml::ixmlstream & xin) {
  xml::attributemap attributes;
  xin.readAttributes(attributes);
  address = attributes["address"];
  xin.readText(content);
  xin.expectNode(xml::endtag());
}












